/*
 * Copyright (c) 2023–2024 Emma Tebibyte <emma@tebibyte.media>
 * SPDX-License-Identifier: AGPL-3.0-or-later
 *
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU Affero General Public License as published by the Free
 * Software Foundation, either version 3 of the License, or (at your option) any
 * later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU Affero General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program. If not, see https://www.gnu.org/licenses/.
 */

use std::{
	env::args,
	io::{ Read, stdin, stdout, Write },
	process::{ Command, exit, Stdio },
};

extern crate getopt;
extern crate strerror;
extern crate sysexits;

use getopt::GetOpt;
use strerror::StrError;
use sysexits::{ EX_DATAERR, EX_IOERR, EX_UNAVAILABLE, EX_USAGE };

fn main() {
	let argv = args().collect::<Vec<String>>();
	let mut d = '\u{1E}'.to_string();
	let mut index_arg = 0;

	let usage = format!(
		"Usage: {} [-d delimiter] index command [args...]",
		argv[0],
	);

	while let Some(opt) = argv.getopt("d:") {
		match opt.opt() {
			Ok(_) => {
				/* unwrap because Err(OptError::MissingArg) will be returned if
				 * opt.arg() is None */
				d = opt.arg().unwrap();
				index_arg = opt.ind();
			},
			Err(_) => {
				eprintln!("{}", usage);
				exit(EX_USAGE);
			}
		};
	}

	let command_arg = index_arg as usize + 1;

	argv.get(command_arg).unwrap_or_else(|| {
		eprintln!("{}", usage);
		exit(EX_USAGE);
	});

	let index = argv[index_arg].parse::<usize>().unwrap_or_else(|e| {
		eprintln!("{}: {}: {}", argv[0], argv[1], e);
		exit(EX_DATAERR);
	});

	let mut buf = String::new();
	let _ = stdin().read_to_string(&mut buf);
	let mut fields = buf.split(&d).collect::<Vec<&str>>();

	let opts = argv
		.iter()
		.clone()
		.skip(command_arg + 1)
		.collect::<Vec<&String>>();

	let mut spawned = Command::new(argv.get(command_arg).unwrap())
		.args(opts)
		.stdin(Stdio::piped())
		.stdout(Stdio::piped())
		.spawn()
		.unwrap_or_else( |e| {
			eprintln!("{}: {}: {}", argv[0], argv[command_arg], e.strerror());
			exit(EX_UNAVAILABLE);
		});

	let field = fields.get(index).unwrap_or_else(|| {
		eprintln!(
			"{}: {}: No such index in input",
			argv[0],
			index.to_string(),
		);
		exit(EX_DATAERR);
	});

	if let Some(mut child_stdin) = spawned.stdin.take() {
		let _ = child_stdin.write_all(field.as_bytes());
		drop(child_stdin);
	}

	let output = spawned.wait_with_output().unwrap_or_else(|e| {
		eprintln!("{}: {}: {}", argv[0], argv[command_arg], e.strerror());
		exit(EX_IOERR);
	});

	let mut replace = output.stdout.clone();

	if replace.pop() != Some(b'\n') { replace = output.stdout; }

	let new_field = String::from_utf8(replace).unwrap_or_else(|e| {
		eprintln!("{}: {}: {}", argv[0], argv[command_arg], e);
		exit(EX_IOERR);
	});

	fields[index] = &new_field;

	stdout().write_all(
		fields.join(&d.to_string()).as_bytes()
	).unwrap_or_else(|e| {
		eprintln!("{}: {}", argv[0], e.strerror());
		exit(EX_IOERR);
	});
}
